;**********************************************************************
;
;    Filename:	    pbus.asm
;    Description:   rs485-type multi-drop bus
;                   with half-duplex serial protocol
;                   one master, max 15 slaves
;                   inv. logic: 0V=+12V=bit0 (START)
;                                 5V=-12V=bit1 (STOP)
;
;    Author:        el@jap.hu
;                   http://jap.hu/electronic/
;
;**********************************************************************
; HISTORY
; 010 - 20010223 initial version
; 011 - 20010224 first working ver, no devid check, rx only
; 012 - 20010224-2 working
; 013 - 20010224-3 respond to cmd 5a with resp 80
; 014 - 20010224-4 ping, version, stats, reset commands implemented
; 015 - 20010307 delayed response (2 bit), external cmd handler
; 016 - 20010516 txport_shadow bugfix & pbus_cmd_handler optimization to fit page0
; 017 - 20010516 only handle 58-5f commands internally
;		 add 5b as CLAST - repeat last response (in the transmit buffer)
;		 add 58 as NOOP - do nothing, just say OK
; 018 - 20010519
; 019 - 20010616
; 040 - 20010624 fix cpingcopy bug, add KEEP_STATS
; 041 - 20010624 modify swusart TX: use TRIS
;**********************************************************************
;
;    Notes:
;
; To use the interrupt handler, SAVE these registers:
; W, STATUS, FSR (if used in main program!)
;
; set and never change the TX portbit from LOW in the main program
;
;**********************************************************************
#define KEEP_STATS

	GLOBAL pbus_init, pbus_int_handler

#ifdef KEEP_STATS
	GLOBAL rxe1cnt, rxe2cnt, rxe3cnt
#endif

	EXTERN pbus_handler

	list      p=16F628     ; list directive to define processor

#include <p16F628.inc>         ; processor specific variable definitions

#define ROK 0x60
#define RECHO 0x6f

; TX definition
; to invert TX, check OPTION bit#6 too!

#define txbit TRISA, 4

; bit1 or STOP (tri-state)
#define TX_1	bsf txbit
; bit0 or START
#define TX_0	bcf txbit

rxport	EQU	PORTB
; rxbit is always bit#0 (it triggers the itr)
; to invert rx, check the copy from rxreg to rxbuf

#define DBIT	PORTB, 3

.serda	UDATA
;***** VARIABLE DEFINITIONS

rxreg	res 1
txreg	res 1
txcnt	res 1 ; tx bit counter
flags	res 1 ; serial routine flags
tcnt	res 1 ; time counter : increase every halfbit time
chksum	res 1 ; buffer checksum

#ifdef KEEP_STATS
rxe1cnt	res 2 ; rx checksum error counter
rxe2cnt	res 2 ; rx all packets counter
rxe3cnt	res 2 ; rx my packets counter
#endif

; don't change buffer order!
rxbuf	res 8 ; receive buffer
txbuf	res 8 ; transmit buffer
rmax	equ 8
tmax	equ 8

#define rxflg flags, 1
#define trflg flags, 2

; * CONSTANTS

#define PBUS_VER 0x41
#define MAIN_VER 0x10
devid		equ 0x30		; lower 4 bits are 0
					; devid 0 is MASTER
halfbit		equ .52			; 208 usec @ 4Mhz, 2400 baud
mingap		equ .20			; min gap between packets by halfbit time since last stop

tmrdelay	equ .256-(2*halfbit)

.pbuslib	CODE

pbus_cmd_vectors
		addlw pbus_cmd
		movwf PCL
pbus_cmd	; command jump table
		goto cnoop		; 8: noop
		goto rxb_ena		; 9: illegal
		goto rxb_ena		; a: illegal
		goto clast		; b: clast

#ifdef KEEP_STATS
		goto cres		; c: cres
		goto cstat		; d: cstat
#else
		goto rxb_ena		; c: illegal
		goto rxb_ena		; d: illegal
#endif

		goto cver		; e: cver
		goto cping		; f: cping

pbus_init	bsf STATUS, RP0 ; select BANK1
		movlw 0x80
		andwf OPTION_REG, W
		iorlw 0x01 ; TMR0 prescaler 1:4, falling edge itr (b#6)
		clrwdt
		movwf OPTION_REG

		TX_1		; tri-state PBUS by default
		bcf STATUS, RP0 ; select BANK0
		clrf flags

#ifdef KEEP_STATS
		clrf rxe1cnt
		clrf rxe1cnt+1
		clrf rxe2cnt
		clrf rxe2cnt+1
		clrf rxe3cnt
		clrf rxe3cnt+1
#endif

		movlw 1
		movwf txcnt
		movlw rxbuf
		movwf FSR ; init rxbuf and enable rx
		bsf rxflg
		clrf tcnt ; pre-packet delay counter
		clrf chksum

		movlw tmrdelay
		movwf TMR0
		movlw (1<<GIE)+(1<<INTE)+(1<<T0IE)
		movwf INTCON
		return

pbus_int_handler; interrupt handler

		btfsc INTCON, INTE ; external int enabled?
		btfss INTCON, INTF ; skip if not triggered
		goto skip_1

rx_trigger	; start receiving a new byte
		movlw .256-(3*halfbit)
		movwf TMR0
		movlw 0x80
		movwf rxreg
		bcf INTCON, INTE	; disable int trigger
		bsf rxflg		; byte rx in progress

		; test for the silence gap
		; if it will be a new packet or not
		; and init FSR if needed
		movlw mingap
		subwf tcnt, W
		btfss STATUS, C
		goto rxb_cont		; if (tcnt < mingap) goto cont

		btfss STATUS, C
rxb_ena		bsf INTCON, INTE

rxb_start	; start receiving a new buffer
		movlw rxbuf
		movwf FSR
		clrf chksum
		bcf trflg

bsf DBIT
nop
nop
nop
nop
bcf DBIT

rxb_cont	bcf INTCON, INTF
		bcf INTCON, T0IF
		return

skip_1		btfss INTCON, T0IF
		return

timer_itr	movlw .256-(2*halfbit)
		movwf TMR0
		bcf INTCON, T0IF

		btfss tcnt, 7		; increase up to 0x80
		incf tcnt, F		; time counter for RX

		; check if TX or RX is in progress
		btfss trflg
		goto chkrx

tx		
		decfsz txcnt,W		; is tx in progress (txcnt > 1 = yes)
		goto chktxcnt		; yes

					; check for a next byte to TX
		movlw 0x0f
		andwf txbuf, W		; length of packet
		addlw (txbuf+2)		; test for packet length
		subwf FSR, W		;

		btfss STATUS, C		; packet sent?
		goto skip_7

txb_done	btfss STATUS, Z		; when equals, TX the checksum
		goto rxb_ena		; else tx done, restart RX

txb_chksum	movf chksum, W
		clrf INDF
		subwf INDF, F		; store & transmit the checksum

skip_7		movf INDF, W		; transfer the char to the TX reg
		incf FSR, F
		movwf txreg
		addwf chksum, F

		movlw .10		; set the TX bit counter
		movwf txcnt
		clrc			; start bit = 0
bsf DBIT
		goto outputbit		; output the start bit

chktxcnt	movwf txcnt		; is it time to output the next bit?

		setc			; stop bit = 1
		rrf txreg,F		; LSB is transmitted first so shift right

bcf DBIT

outputbit	bsf STATUS, RP0 ; select BANK1
		skpc
		TX_0			; no carry
		skpnc
		TX_1			; carry
		bcf STATUS, RP0 ; select BANK0
		return

chkrx		btfss rxflg
		return

rx		rrf rxport,W		; move Rx Bit into carry
		rrf rxreg,F		; rotate carry into rxreg LSB first
		skpc			; if carry set, RX is done
		return

		; receive byte is done: check if there are more
		; and need to enable ext. int again

		bcf rxflg		; byte rx is done
		clrf tcnt		; gap counter between bytes
		movf rxreg, W
		movwf INDF
		addwf chksum, F
		incf FSR, F

		movlw 0x0f
		andwf rxbuf, W		; length of packet
		addlw (rxbuf+3)		; test for packet end
		subwf FSR, W		; test for overflow!!!

		btfsc STATUS, Z		; packet received?
		goto rxb_done
		
		bcf INTCON, INTF	; no, re-enable the rx interrupt
		bsf INTCON, INTE
		return

rxb_done	; packed received, don't re-enable rx (INTE)

		movf chksum, W		; check packet
		btfsc STATUS, Z
		goto skip_6

#ifdef KEEP_STATS
		incf rxe1cnt, F		; packet checksum error
		btfsc STATUS, Z
		incf rxe1cnt+1, F
#endif

		goto rxb_ena		; restart with new packet

skip_6
#ifdef KEEP_STATS
		incf rxe2cnt, F		; count good pakets
		btfsc STATUS, Z
		incf rxe2cnt+1, F
#endif

		movlw 0xf0		; check devid
		andwf rxbuf, W
		sublw devid
		btfss STATUS, Z
		goto rxb_ena		; devid mismatch, restart

;************** Command handlers start here
; illegal commands with no response goto rxb_ena (restart with a new packet)
; good commands with resp. in txbuf goto txb_ena (TX a packet with checksum)
;**************
handle_cmd	; got a good packet

#ifdef KEEP_STATS
		incf rxe3cnt, F
		btfsc STATUS, Z
		incf rxe3cnt+1, F
#endif

		movlw 0xf8		; check for higher 5 bits
		andwf (rxbuf+1), W
		sublw 0x58
		btfss STATUS, Z
		goto cext		; not internal

		movlw 0x07
		andwf (rxbuf+1), W
		goto pbus_cmd_vectors

cping		movlw RECHO
		movwf txbuf+1

		movlw 0x0f
		andwf rxbuf, W
		movwf txbuf
		movwf chksum		; use as counter
		bz txb_ena		; data present?

		movlw rxbuf+2		; copy data to response
		movwf FSR

cpingcopy	movf INDF, W
		movwf rxbuf+1		; use as temp storage
		movlw rmax
		addwf FSR, F		; goto transmit buffer
		movf rxbuf+1, W
		movwf INDF

		movlw rmax
		subwf FSR, F		; goto receive buffer
		incf FSR, F
		decfsz chksum, F
		goto cpingcopy

		goto txb_ena

cver		movlw 2
		movwf txbuf
		movlw ROK
		movwf txbuf+1
		movlw PBUS_VER
		movwf txbuf+2
		movlw MAIN_VER
		movwf txbuf+3
		goto txb_ena

#ifdef KEEP_STATS
cstat		movlw 6
		movwf txbuf
		movlw ROK
		movwf txbuf+1
		movf rxe1cnt, W
		movwf txbuf+2
		movf rxe1cnt+1, W
		movwf txbuf+3
		movf rxe2cnt, W
		movwf txbuf+4
		movf rxe2cnt+1, W
		movwf txbuf+5
		movf rxe3cnt, W
		movwf txbuf+6
		movf rxe3cnt+1, W
		movwf txbuf+7
		goto txb_ena

cres		clrf rxe1cnt
		clrf rxe1cnt+1
		clrf rxe2cnt
		clrf rxe2cnt+1
		clrf rxe3cnt
		clrf rxe3cnt+1
#endif

cnoop		movlw 0
		movwf txbuf
		movlw ROK
		movwf rxbuf+1
		goto txb_ena

cext		movf (rxbuf+1), W
		call pbus_handler
		andlw 0xff		; if at return W=0, respond
		btfss STATUS, Z
		goto rxb_ena		; otherwise, go back listening

clast		; repeat last response (in the transmit buffer)
txb_ena		; start a TX cycle
		movlw txbuf
		movwf FSR
		clrf chksum
		bsf trflg		; set transmit mode
		movlw 2
		movwf txcnt
		movlw 0xff
		movwf txreg		;; tx 2 bit1-s before packet
		return

	end

